from socket import AF_INET, SOCK_STREAM, SHUT_RDWR
from socket import socket as Socket
from socket import create_connection

from typing import Union

from .exceptions import InvalidRequest, InvalidResponse



class Memory:
    """Client for reading from and writing to Bizhawk memory"""

    def __init__(self, domain: str, address: str='127.0.0.1', port: int=16154):
        self.domain = domain
        
        self.address = address
        self.port = port


    def _request(self, query: Union[bytes, str]):
        """Send a request to the Bizhawk Lua memory hook


        Pattern:  DOMN/ADDR/TSLE/VLUE

        DOMN  - Memory domain
        ADDR  - Address
        TSLE--.-------------------.
            Type:               Signage:
            | * [b]yte            * [u]nsigned
            | * [i]nteger         * [s]igned
            | * [f]loat
            |
            .-------------------.
            Length              Endianness
                * [1] byte          * [l]ittle endian
                * [2] bytes         * [b]ig endian
                * [3] bytes
                * [4] bytes

        VLUE  - Integer or float, ex. 12 or -1.2
        

        # EXAMPLES

        VRAM/11423051/iu4b/23
            Write (23) to (11423051) in (VRAM)
            an [i]nteger, [u]nsigned,
            [4] bytes long and [b]ig endian

        WRAM/21512962/fs2l/
            Read from (21512962) in (WRAM)
            a [f]loat, [s]igned,
            [2] bytes long and [l]ittle endian
        """

        # Socket requires a byte string to send
        if type(query) is not bytes:
            query = query.encode()


        with create_connection((self.address, self.port)) as socket:
            # Send request and expect response
            socket.sendall(query)
            response = self._receive(socket)


        try:
            # Extract response code and message
            code, _, message = response.decode('UTF-8').partition('_')
            code = int(code)
        
        except ValueError:
            raise InvalidResponse('Response could not be divided into code and message')


        # Successfully wrote to memory
        if code == 0:
            return True

        # Successfully read byte
        if code == 1:
            return response[response.index(b'_'):]

        # Successfully read integer
        if code == 2:
            return int(message)

        # Successfully read float
        if code == 3:
            return float(message)


        raise InvalidRequest(code, message)

    def _receive(self, socket: Socket, n: int=128):
        """Receive data until end of stream"""

        # Receive data in packets
        # and concatenate them at the end

        buffer = []

        while True:
            data = socket.recv(n)

            if not data:
                break

            buffer.append(data)

        return b''.join(buffer)

    def _format_tsle(self, type_: type, signed: bool, length: int, endianness: str):
        """Format the type, signage, length, and endianness for a request"""

        if type_ not in (bytes, int, float):
            raise ValueError('Type must be bytes, int or float')

        if length not in (1, 2, 3, 4):
            raise ValueError('Length must be 1, 2, 3 or 4 bytes')

        if endianness not in ('little', 'big'):
            raise ValueError('Endianness must be little or big')


        return ''.join([
            type_.__name__[0],
            'us'[signed],
            str(length),
            endianness[0]
        ])
    
    def _format_query(self, address: int, type_: type=bytes, signed: bool=False,
                        length: int=1, endianness: str='big', value: Union[int, float]=None):
        """Format all request parameters into a valid query for a request"""

        tsle = self._format_tsle(type_, signed, length, endianness)            
        return f'{self.domain}/{address}/{tsle}/{"" if value is None else value}'


    def read_byte(self, address: int):
        """Read byte from memory"""
        return self._request(self._format_query(
            address=address,
            type_=bytes,
            value=None
        ))

    def write_byte(self, address: int, value: int):
        """Write byte from memory"""
        return self._request(self._format_query(
            address=address,
            type_=bytes,
            value=value
        ))

    def read_int(self, address: int, signed: bool=False, length: int=1, endianness: str='big'):
        """Read integer from memory"""
        return self._request(self._format_query(
            address=address,
            type_=int,
            signed=signed,
            length=length,
            endianness=endianness,
            value=None
        ))

    def write_int(self, address: int, value: int, signed: bool=False, length: int=1, endianness: str='big'):
        """Write integer from memory"""
        return self._request(self._format_query(
            address=address,
            type_=int,
            signed=signed,
            length=length,
            endianness=endianness,
            value=value
        ))

    def read_float(self, address: int, endianness: str='big'):
        """Read float from memory"""
        return self._request(self._format_query(
            address=address,
            type_=float,
            endianness=endianness,
            value=None
        ))

    def write_float(self, address: int, value: int, endianness: str='big'):
        """Write float from memory"""
        return self._request(self._format_query(
            address=address,
            type_=float,
            endianness=endianness,
            value=value
        ))


import tkinter as tk
from tkinter import filedialog

from pathlib import Path
from zipfile import ZipFile



def export_lua_components(path: str=None):
    """Export the Lua components for Bizhawk"""

    if path is None:
        # Open a dialog window for the user
        # to specify where to export the files
        root = tk.Tk()
        root.withdraw()  # Hide the root window
        path = filedialog.askdirectory()

    # Get the path to where the ZIP file is stored
    from_ = str(Path(__file__).parent / 'lua_components.zip')

    # Extract!
    with ZipFile(from_, 'r') as f:
        f.extractall(path)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bizhook']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'bizhook',
    'version': '1.0.0',
    'description': 'Communicate with the Bizhawk emulator via a Lua socket',
    'long_description': "# Bizhawk Hook\n\nInteract with [Bizhawk](http://tasvideos.org/Bizhawk.html) via a socket server hosted in the Lua Console.\n\n## How to use\n\n### Bizhawk Lua\n\n#### Exporting\n\nExport all the necessary Lua components using the provided function.\n```py\nfrom bizhook import export_lua_components\n\nexport_lua_components('/home/williamson/.bizhook')\n```\nYou can either provide a path or leave it empty to have it open up a file dialogue asking for directory.\n\n#### Opening socket\n\nIn Bizhawk, go to `Tools` > `Lua Console`. Select `Open script` and open `hook.lua` from the exported components.\n\n##### Is it working?\n\nIf it starts successfully, you should see a text in the top-left of the emulator saying the socket is being opened. Should that not appear, try restarting the emulator until it does. This seems to be an issue with Bizhawk.\n\n**Note**: Do not try to communicate with the socket *before* the text has disappeared, as it isn't actually opened yet. The message is there to make it clear that the script is running successfully.\n\n### Python\n\nYou can read from and write to memory by using a `Memory` object.\n```py\nfrom bizhook import Memory\n\ncombined_wram = Memory('Combined WRAM')\n```\n\nTo see the available methods, do `help(Memory)`.\n\n#### Memory domain\nYou can use the default memory domain by providing an empty string. However, I would advice against it and that you always do specify with which domain you want to interact. It may be that it works for you solely because, per chance, the default one happens to be the correct one.",
    'author': 'Maximillian Strand',
    'author_email': 'maximillian.strand@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/thedisruptproject/bizhook',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)




class BizhawkHookError(Exception):
    """Base error for Bizhawk hook"""


class InvalidRequest(BizhawkHookError):
    """Bizhawk was sent an invalid request"""

    def __init__(self, code, message):
        super().__init__(message)
        self.code = code

class InvalidResponse(BizhawkHookError):
    """Response did not follow the parseable format"""